/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.bytesw.bytebot.service;

import com.bytesw.bytebot.http.FileValidationResponse;
import com.bytesw.bytebot.http.FileValidationResult;
import com.bytesw.bytebot.http.enums.ValidationStatusEnum;
import com.bytesw.bytebot.model.QuestionFile;
import com.bytesw.bytebot.model.enums.FrequentQuestionStatusEnum;
import com.bytesw.bytebot.repository.QuestionFileRepository;
import com.bytesw.bytebot.utils.Utilities;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.io.FilenameUtils;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellType;
import org.apache.poi.ss.usermodel.DateUtil;
import org.apache.poi.ss.usermodel.RichTextString;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.WorkbookFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

@Service
@Transactional
@Log4j2
public class FileManagementService {

    String[] validExtensions = new String[]{"xls", "xlsx"};
    String[] fileHeaders = new String[]{"TOPICO", "SUBTOPICO", "PREGUNTA", "RESPUESTA"};
    
    @Autowired
    private QuestionFileRepository questionFileRepository;
    
    @Transactional(propagation = Propagation.REQUIRED)
    public FileValidationResponse validateAndSaveFile(String uuid, MultipartFile file) throws IOException {

        FileValidationResponse response = new FileValidationResponse();
        FileValidationResult result = new FileValidationResult();

        //@Validaciones: Tipo de Extension
        boolean acceptedExtension = false;
        String fileExtension = FilenameUtils.getExtension(file.getOriginalFilename());
        for (String validExension : validExtensions) {
            if (validExension.equalsIgnoreCase(fileExtension.trim())) {
                acceptedExtension = true;
                continue;
            }
        }

        if (!acceptedExtension) {
            log.info("La extension [" + fileExtension + "] es inválida");
            result.setStatus(ValidationStatusEnum.INCOMPATIBLE_EXTENSION);
            response.setFileValidationResult(result);
            return response;
        }
        
        //@Validaciones: De Estructura del Excel (Cabeceras)
        byte[] fileData = file.getBytes();
        log.debug("File Size: [" + file.getSize() + "]");
        ByteArrayInputStream byteArrayInputStream = new ByteArrayInputStream(fileData);
        Workbook workbook = WorkbookFactory.create(byteArrayInputStream);
        Sheet sheet = workbook.getSheetAt(0);
        Row headerRow = sheet.getRow(0);

        //@HU: Llenando los headers encontrados
        Map<String, Integer> headersXPosition = new HashMap<>();
        List<String> headersList = new ArrayList<>();
        
        Iterator<Cell> cells = headerRow.cellIterator();
        int position = 0;
        while (cells.hasNext()) {
            Cell cell = (Cell) cells.next();
            RichTextString value = cell.getRichStringCellValue();
            headersXPosition.put(value.getString(), position++);
            headersList.add(value.getString());
        }

        Map<String, String> headerErrorMap = validateHeaders(headersList);

        if (!headerErrorMap.isEmpty()) {
            log.info("Se encontraron los siguientes errores: " + headerErrorMap.values());
            result.setStatus(ValidationStatusEnum.HEADER_ERROR);
            result.setHeadersErrorMap(headerErrorMap);
            response.setFileValidationResult(result);
            return response;
        }

        //@Validaciones: De Contenido(Todos deben tener valor)
        int rowNumber = 0;
        Iterator<Row> rowIterator = sheet.iterator();
        Map<Integer, Object> recordsRowErrorMap = new HashMap<>();;

        while (rowIterator.hasNext()) {
            Row row = (Row) rowIterator.next();
            if (rowNumber == 0) {
                rowNumber++;
                continue;
            }

            log.debug("Row number: [" + rowNumber + "]");
            Iterator<String> columnsIterator = headersXPosition.keySet().iterator();
            Map<String, Object> recordsXRow = new HashMap<>();

            while (columnsIterator.hasNext()) {
                String name = columnsIterator.next();
                log.debug("Read columnName [" + name + "]");

                if (!isFileHeader(name.toUpperCase().trim())) {
                    log.debug("Skip columnName [" + name + "] - not required in format");
                    continue;
                }

                Cell cell = row.getCell(headersXPosition.get(name));

                if (cell == null) {
                    recordsXRow.put(name, null);
                    continue;
                }

                CellType cellType = row.getCell(headersXPosition.get(name)).getCellType();

                switch (cellType) {
                    case STRING:
                        RichTextString cellValue = row.getCell(headersXPosition.get(name)).getRichStringCellValue();
                        String value = (cellValue != null ? cellValue.getString().trim() : null);
                        recordsXRow.put(name, value);
                        break;

                    case NUMERIC:
                        if (DateUtil.isCellDateFormatted(row.getCell(headersXPosition.get(name)))) {
                            Date date = row.getCell(headersXPosition.get(name)).getDateCellValue();
                            recordsXRow.put(name, Utilities.convertDateToString(date, "dd-MMM-yyyy"));
                        } else {
                            double numericVal = row.getCell(headersXPosition.get(name)).getNumericCellValue();
                            recordsXRow.put(name, numericVal);
                        }
                        break;

                    case BLANK:
                        recordsXRow.put(name, null);
                        break;

                    default:
                        log.info(" CellType not supported yet [" + row.getCell(headersXPosition.get(name)).getCellType() + "]");
                        log.info(" Value [" + row.getCell(headersXPosition.get(name)).getStringCellValue() + "]");
                }
            }

            Map<String, String> recordsErrorMap = validateRecordsXRow(recordsXRow);

            if (!recordsErrorMap.isEmpty()) {
                recordsRowErrorMap.put(rowNumber + 1, recordsErrorMap);
            }

            rowNumber++;
        }

        if (!recordsRowErrorMap.isEmpty()) {
            log.info("Hay errores en el contenido del archivo: " + recordsRowErrorMap);
            result.setStatus(ValidationStatusEnum.CONTENT_ERROR);
            result.setRecordsErrorMap(recordsRowErrorMap);
            response.setFileValidationResult(result);
            return response;
        }

        //@HU: Proceder a guardar en BD el archivo subido
        result.setStatus(ValidationStatusEnum.OK);
        
        QuestionFile questionFileBD = new QuestionFile();
        
        questionFileBD.setUuid(uuid);
        questionFileBD.setName(file.getOriginalFilename());
        questionFileBD.setSize(file.getSize());
        questionFileBD.setUploadDate(LocalDateTime.now());
        questionFileBD.setData(fileData);
        
        questionFileBD = questionFileRepository.save(questionFileBD);
        
        response.setId(questionFileBD.getId());
        response.setUuid(questionFileBD.getUuid());
        response.setFileName(questionFileBD.getName());
        response.setStatus(FrequentQuestionStatusEnum.PENDING_SYNCHRONIZED.getName());
        response.setUploadDate(getUploadDate(questionFileBD.getUploadDate()));
        
        response.setFileValidationResult(result);
        
        return response;
    }

    private String getUploadDate(LocalDateTime uploadDate) {
        if (uploadDate != null) {
            return uploadDate.format(DateTimeFormatter.ofPattern("dd/MM/yyyy"));
        }
        return null;
    }

    private Map<String, String> validateHeaders(List<String> headersList) {

        Map<String, Integer> headerOcurrencesMap = new HashMap<>();
        Map<String, String> headerErrorMap = new HashMap<>();

        for (String fileHeader : fileHeaders) {
            String keyHeader = fileHeader.toUpperCase().trim();
            if (headerOcurrencesMap.get(keyHeader) == null) {
                headerOcurrencesMap.put(keyHeader, 0);
            }

            for (String header : headersList) {
                if (header.toUpperCase().trim().equals(keyHeader)) {
                    int ocurrences = headerOcurrencesMap.get(keyHeader);
                    headerOcurrencesMap.put(keyHeader, ocurrences + 1);
                }
            }
        }

        for (String keyHeader : headerOcurrencesMap.keySet()) {
            int ocurrences = headerOcurrencesMap.get(keyHeader);
            if (ocurrences == 0) {
                headerErrorMap.put(keyHeader, "HEADER.NOT.FOUND");
            } else if (ocurrences > 1) {
                headerErrorMap.put(keyHeader, "HEADER.REPETEAD");
            }
        }
        return headerErrorMap;
    }

    private Map<String, String> validateRecordsXRow(Map<String, Object> recordsXRow) {
        Map<String, String> recordsErrorMap = new HashMap<>();

        for (String headerName : recordsXRow.keySet()) {

            String keyHeader = headerName.toUpperCase().trim();
            if (!isFileHeader(keyHeader)) {
                continue;
            }
            if (recordsErrorMap.get(keyHeader) != null) {
                continue;
            }

            if (recordsXRow.get(headerName) == null) {
                recordsErrorMap.put(keyHeader, "NOT.FOUND.VALUE");
            } else {
                Object recordRowValue = recordsXRow.get(headerName);

                if (recordRowValue instanceof String) {
                    String record = (String) recordRowValue;
                    if (record.trim().equals("")) {
                        recordsErrorMap.put(keyHeader, "EMPTY.VALUE");
                    }
                }
            }
        }

        return recordsErrorMap;
    }

    private boolean isFileHeader(String keyHeader) {
        boolean isFileHeader = false;
        for (String fileHeader : fileHeaders) {
            if (keyHeader.equals(fileHeader.toUpperCase().trim())) {
                isFileHeader = true;
                break;
            }
        }
        return isFileHeader;
    }
}
